/* -*- mode: c++; c-basic-offset: 2; -*- */

#include "general/config.h"

#include <signal.h>
#include <pthread.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include "general/unordered_map.hh"
#include "general/ldasexception.hh"
#include "general/Thread.hh"

using General::Thread;

static General::Thread::thread_type NULL_THREAD =
  (General::Thread::thread_type)0;

class thread_registry
{
public:
  typedef Thread::thread_type	key_type;
  typedef Thread* value_type;

  static void Add( key_type, value_type );

  static value_type Get( key_type );

private:
  typedef General::unordered_map< key_type, value_type > thread_registry_type;

  static thread_registry_type	m_storage;
  static MutexLock::lock_type	m_baton;
};

void thread_registry::
Add( thread_registry::key_type Key, thread_registry::value_type Value )
{
  MutexLock	l( m_baton);

  m_storage[ Key ] = Value;
}

thread_registry::value_type thread_registry::
Get( thread_registry::key_type Key )
{
  MutexLock	l( m_baton);

  thread_registry_type::const_iterator pos( m_storage.find( Key ) );
  if ( pos == m_storage.end( ) )
  {
    std::ostringstream	msg;
    msg << "There is no managed thread with key: " << (void*)Key;

    throw std::range_error( msg.str( ) );
  }
  return pos->second;
}

thread_registry::thread_registry_type thread_registry::m_storage;
MutexLock::lock_type thread_registry::m_baton = MutexLock::Initialize( );

namespace General
{
  size_t Thread::m_stack_size_default( PTHREAD_STACK_MIN );

  Thread::
  Thread( )
    : m_parent_tid( pthread_self( ) ),
      m_tid( NULL_THREAD ),
      m_key( MutexLock::Initialize( ) ),
      m_cancel_via_exception( false ),
      m_cancel_state( false ),
      m_cancel_thread( false )
  {
  }

  Thread::
  ~Thread( )
  {
    Join( );
  }

  bool Thread::
  operator==( const Thread& Source ) const
  {
    return( Source.m_tid == m_tid );
  }

  void Thread::
  Cancel( )
  {
    if ( m_tid != NULL_THREAD )
    {
      if ( m_cancel_via_exception )
      {
	m_cancel_thread = true;
      }
      else
      {
	pthread_cancel( m_tid );
      }
    }
  }

  //---------------------------------------------------------------------
  /// Deliver any pending cancellation requests to the calling thread.
  /// If cancellation is to be done via exception,
  /// then a cancellation exception is throw.
  /// If not, then a call to the appropriate thread library's
  /// cancellation routine is made.
  ///
  /// \note
  ///     This call must only be made by a child thread.
  //---------------------------------------------------------------------
  void Thread::
  CancellationCheck( const std::string& Header,
		     const char* File, const int Line ) const
  {
#if 1
    if ( ! pthread_equal( m_tid, pthread_self( ) ) )
    {
      return;
    }
    if ( ( m_cancel_via_exception )
	 && ( m_cancel_thread ) )
    {
      throw cancellation( Header, File, Line );
    }
    pthread_testcancel( );
#endif /* 0 */
  }

  //---------------------------------------------------------------------
  /// Establish the cancellability of a thread.
  /// If a Value of true is passed, then the thread will allow
  /// cancellation as determined by the call to CancellationType.
  ///
  /// \note
  ///     This call must only be made by a child thread.
  ///
  /// \see
  ///     CancellationType
  //---------------------------------------------------------------------
  void Thread::
  CancellationEnable( bool Value )
  {
    if ( ! pthread_equal( m_tid, pthread_self( ) ) )
    {
      return;
    }
    int oldstate;

    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );
    m_cancel_state = Value;
    if ( ( m_cancel_state )
	 && ( m_cancel_via_exception == false ) )
    {
      pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, &oldstate );
    }
    
  }

  //---------------------------------------------------------------------
  /// Sets a how a thread can be cancelled.
  /// <ul>
  ///   <li><b>CANCEL_ASYNCHRONOUS</b>
  ///       A cancellation request is immediated delivered to the
  ///       thread.
  ///   <li><b>CANCEL_DEFERRED</b>
  ///       A cancellation request is marked pending for the thread
  ///       and the thread is cancelled when it gets to a system
  ///       cancellation point or when CancellationCheck is called.
  ///   <li><b>CANCEL_EXCEPTION</b>
  ///       A cancellation request is marked pending for the thread
  ///       and the thread throws a cancellation exception upon
  ///       calling CancellationCheck.
  /// </ul>
  ///
  /// \note
  ///     This call must only be made by a child thread.
  ///
  /// \see
  ///     CancellationEnable, CancellationCheck
  //---------------------------------------------------------------------
  void Thread::
  CancellationType( cancel_type Type )
  {
    if ( ! pthread_equal( m_tid, pthread_self( ) ) )
    {
      return;
    }
    int oldtype;
    int oldstate;

    pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &oldstate );
    switch( Type )
    {
    case CANCEL_ASYNCHRONOUS:
      m_cancel_via_exception = false;
      pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype );
      break;
    case CANCEL_DEFERRED:
      m_cancel_via_exception = false;
      pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, &oldtype );
      break;
    case CANCEL_EXCEPTION:
      m_cancel_via_exception = true;
      break;
    }
    if ( m_cancel_via_exception == false )
    {
      // Restore the state of cancellation
      pthread_setcancelstate( oldstate, &oldstate );
    }
  }

  bool Thread::
  IsParentThread( ) const
  {
    return ( pthread_self( ) == m_parent_tid );
  }

  void Thread::
  Detach( ) const
  {
    if ( m_tid != NULL_THREAD )
    {
      // Detach the current thread (makes it non-joinable).
      volatile thread_type tmp = m_tid;

      {
	MutexLock	l( m_key, __FILE__, __LINE__ );

	m_tid = NULL_THREAD;
      }
      // thread_registry( ).erase( m_tid );
      pthread_detach( tmp );
    }
  }

  void Thread::
  Join( )
  {
    if ( m_tid != NULL_THREAD )
    {
      bool release( true );

      int status = pthread_join( m_tid, NULL );
      try
      {
	switch( status )
	{
	case ESRCH:
	  throw range_error( "Thread::Join", __FILE__, __LINE__ );
	  break;
	case EINVAL: 
	  throw invalid_argument( "Thread::Join", __FILE__, __LINE__ );
	  break;
	case EDEADLK: 
	  release = false;
	  throw deadlock( "Thead::Join", __FILE__, __LINE__ );
	  break;
	}
      }
      catch( ... )
      {
	if ( release )
	{
	  // get_thread_registry( ).erase( m_tid );
	  
	  MutexLock	l( m_key, __FILE__, __LINE__ );
	  m_tid = NULL_THREAD;
	}
	throw; // rethrow the exception
      }
      {
	// get_thread_registry( ).erase( m_tid );

	MutexLock	l( m_key, __FILE__, __LINE__ );
	m_tid = NULL_THREAD;
      }
    }
  }

  int Thread::
  Kill( int Signal ) const
  {
    if ( m_tid != NULL_THREAD )
    {
      return pthread_kill( m_tid, Signal );
    }
    // Process could not be found since it is null
    return ESRCH;
  }

  Thread& Thread::
  Self( )
  {
    return *( thread_registry::Get( pthread_self( ) ) );
  }

  
  Thread& Thread::
  Self( thread_type Key )
  {
    return *( thread_registry::Get( Key ) );
  }

  int Thread::
  Spawn( )
  {
    return spawn( );
  }

  int Thread::
  spawn( start_function_type StartFunction )
  {
    int retval = 0;
    if ( m_tid == NULL_THREAD )
    {

      pthread_attr_t attr;
      pthread_attr_init (&attr);
      if ( StackSizeDefault( ) )
      {
	pthread_attr_setstacksize ( &attr, StackSizeDefault( ) );
      }

      retval = pthread_create( &m_tid, &attr, StartFunction, this );
      pthread_attr_destroy (&attr);

      if ( retval != 0 )
      {
	throw std::runtime_error( "General::Thread::Start: failed to start thread" );
      }
      thread_registry::Add( m_tid, this );
    }
    return retval;
  }

  Thread* Thread::
  start_routine( Thread* ThreadSource )
  {
    ThreadSource->action( );

    return ThreadSource;
  }

  Thread::cancellation::
  cancellation( const std::string& Header,
	       const char* File, const int Line )
    : std::runtime_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::cancellation::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": thread cancellation."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::deadlock::
  deadlock( const std::string& Header,
	       const char* File, const int Line )
    : std::logic_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::deadlock::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": attempt to join with self."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::invalid_argument::
  invalid_argument( const std::string& Header,
	       const char* File, const int Line )
    : std::invalid_argument( format( Header, File, Line ) )
  {
  }

  std::string Thread::invalid_argument::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": invalid argument."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }

  Thread::range_error::
  range_error( const std::string& Header,
	       const char* File, const int Line )
    : std::range_error( format( Header, File, Line ) )
  {
  }

  std::string Thread::range_error::
  format( const std::string& Header,
	  const char* File, const int Line )
  {
    std::ostringstream	msg;

    msg << Header << ": no thread could be found."
	<< " (File: " << File << " Line: " << Line << ")"
      ;
    return msg.str( );
  }
}
